#ident	"@(#)slconnect.c	1.5"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stropts.h>
#include <termios.h>
#include <netdb.h>
#include <signal.h>
#include <fcntl.h>
#include <poll.h>
#include <ctype.h>
#include <slipcmd.h>

#define CHATBUFSZ	4096
#define isoctal(X)	(isdigit(X) && ((X) < '8'))
#define tooctal(N, X)	(((N) * 8) + ((X) - '0'))

typedef struct {
	struct termio	*attr;		/* terminal line settings */
	int		baud;		/* unused */
	int		speed;		/* -1 or 0 for any speed TODO: unused? */
	char		*line;		/* device name */
	char		*telno;		/* phone number or system name */
	int		modem;		/* unused */
	char		*device;	/* unused */
	int		dev_len;	/* unused */
} CALL;

/* dial error codes */
#define INTRPT	-1	/* interrupted */
#define D_HUNG	-2	/* dialer hung */
#define NO_ANS	-3	/* no answer */
#define ILL_BD	-4	/* illegal baud rate */
#define A_PROB	-5	/* device open failed */
#define L_PROB	-6	/* device open failed */
#define NO_Ldv	-7	/* can't open Devices file */
#define DV_NT_A	-8	/* device unavailable */
#define DV_NT_K	-9	/* device unknown */
#define NO_BD_A	-10	/* baud rate for device unavailable */
#define NO_BD_K	-11	/* baud rate for device unknown */
#define DV_NT_E	-12	/* speed doesn't match */
#define BAD_SYS	-13	/* unknown system name */
#define CS_PROB	-14	/* connection server error */

static char *cmd;

extern int dial(CALL);
extern void undial(int);
static void dialerr(int);
static void timeout(int);
static void cleanup(int);
static void nop(int);
static int chat(int, char *);
static int parse(char **, char **, char **, char **,
	char **);
static int doexpect(int, const char *);
static int dosend(int, const char *);

void
main(int argc, char *argv[])
{
	int fd, pfd, n, err, server;
	CALL call;
	char *otherhost;
	struct termios t;
	struct conninfo *cip;
	struct slcommand slcmd;
	struct sigaction sa;

	cmd = strrchr(argv[0], '/');
	if (cmd == NULL)
		cmd = argv[0];
	else
		cmd++;
	err = server = 0;
	if (argc == 1) {
		server = 1;
	} else if (argc != 2) {
		fprintf(stderr, "usage: %s [system]\n", cmd);
		exit(1);
	}
	if (server != 0) {
		otherhost = getenv("L0");
		if (otherhost == NULL)
			exit(1);
		close(1);
		close(2);
		if ((open("/dev/null", O_WRONLY) != 1) ||
		  (dup(1) != 2))
			exit(1);
	} else {
		otherhost = argv[1];
	}

	/*
	 * Get the information from /etc/slipsys.
	 */
	cip = getcinfo(otherhost);
	if (cip == NULL)
		fatal("%s: can't get system information", cmd);

	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	if (server == 0) {
		/*
		 * Create a connection to the remote host.
		 */
		sa.sa_handler = cleanup;
		sigaction(SIGINT, &sa, NULL);
		sigaction(SIGQUIT, &sa, NULL);
		sigaction(SIGTERM, &sa, NULL);
		call.telno = otherhost;
		call.attr = NULL;
		call.speed = 0;
		call.line = NULL;
		fd = dial(call);
		if (fd < 0) {
			dialerr(fd);
			exit(1);
		}
	} else {
		fd = 0;
	}

	/*
	 * Set the line characteristics.
	 */
	tcgetattr(fd, &t);
	t.c_cflag &= ~(CSIZE|PARENB|PARODD|CLOCAL|CSTOPB);
	t.c_cflag |= CREAD|HUPCL|CS8;
	t.c_lflag &= ~ECHO;
	tcsetattr(fd, TCSANOW, &t);

	if (server == 0) {
		/*
		 * Complete the connection protocol with the remote host.
		 */
		if (chat(fd, cip->chatscript) < 0) {
			undial(fd);
			fatal("%s: chat failed", cmd);
		}
	}

	/*
	 * Get rid of unnecessary modules.
	 */
	n = ioctl(fd, I_LIST, NULL);
	while (n > 1) {
		ioctl(fd, I_POP, 0);
		n--;
	}

	/*
	 * Prevent the driver from doing more than we want it to.
	 */
	t.c_lflag = 0;
	t.c_iflag = 0;
	t.c_oflag = 0;
	tcsetattr(fd, TCSANOW, &t);

	/*
	 * Now send the request to the slip daemon.
	 */
	slcmd.opcode = CONNECT;
	strcpy(slcmd.arg, otherhost);
	pfd = open(PIPEPATH, O_RDWR);
	if (pfd < 0) {
		if (server == 0)
			undial(fd);
		fatal("%s: can't open pipe to slip daemon", cmd);
	}
	n = write(pfd, &slcmd, sizeof(struct slcommand));
	if (n != sizeof(struct slcommand)) {
		if (server == 0)
			undial(fd);
		if (n < 0)
			fatal("%s: can't write to slipd pipe", cmd);
		else
			fatal("%s: short write to slipd pipe", cmd);
	}
	if (ioctl(pfd, I_SENDFD, fd) < 0) {
		if (server == 0)
			undial(fd);
		fatal("%s: can't send fd to slip daemon", cmd);
	}
	sa.sa_handler = timeout;
	sigaction(SIGALRM, &sa, NULL);
	alarm(30);
	n = read(pfd, &err, sizeof(int));
	alarm(0);
	if (server == 0)
		undial(fd);	/* remove uucp lock file */
	if (n != sizeof(int)) {
		if (n < 0)
			fatal("%s: can't read from slipd pipe", cmd);
		else
			fatal("%s: short read from slipd pipe", cmd);
	}
	if (err == 0)
		printf("connection established\n");
	else
		fatal("%s: no connection: %s", cmd, strerror(err));
	exit(0);
}

static void
dialerr(int code)
{
	switch (code) {
	case INTRPT:
		fprintf(stderr, "%s: call interrupted\n", cmd);
		break;

	case D_HUNG:
		fprintf(stderr, "%s: call hung\n", cmd);
		break;

	case NO_ANS:
		fprintf(stderr, "%s: no answer\n", cmd);
		break;

	case ILL_BD:
		fprintf(stderr, "%s: invalid baud rate\n", cmd);
		break;

	case A_PROB:
	case L_PROB:
		fprintf(stderr, "%s: can't open device\n", cmd);
		break;

	case NO_Ldv:
		fprintf(stderr, "%s: can't open uucp files\n", cmd);
		break;

	case DV_NT_A:
		fprintf(stderr, "%s: device in use\n", cmd);
		break;

	case DV_NT_K:
	case NO_BD_K:
		fprintf(stderr, "%s: device not found\n", cmd);
		break;

	case NO_BD_A:
		fprintf(stderr, "%s: no available devices\n", cmd);
		break;

	case DV_NT_E:
		fprintf(stderr, "%s: unsupported speed\n", cmd);
		break;

	case BAD_SYS:
		fprintf(stderr, "%s: unknown system\n", cmd);
		break;

	case CS_PROB:
		fprintf(stderr, "%s: can't reach connection server\n",
		  cmd);
		break;

	default:
		fprintf(stderr, "%s: dial error (%d)\n", cmd, code);
	}
}

static void
timeout(int signo)
{
	undial(-1);
	fatal("%s: request timed out", cmd);
}

static void
cleanup(int signo)
{
	undial(-1);
	fatal("%s: interrupt (signal %d)", cmd, signo);
}

static int
chat(int fd, char *script)
{
	char *expectstr, *sendstr, *subexpectstr, *subsendstr;
	struct sigaction sa;

	sigemptyset(&sa.sa_mask);
	sa.sa_handler = nop;
	sa.sa_flags = 0;
	sigaction(SIGALRM, &sa, NULL);
	while (parse(&script, &expectstr, &sendstr, &subexpectstr,
	  &subsendstr)) {
		if (doexpect(fd, expectstr)) {
			if (dosend(fd, sendstr) < 0)
				return(-1);
		} else if (subexpectstr) {
			if (dosend(fd, subsendstr) < 0)
				return(-1);
			if (doexpect(fd, subexpectstr)) {
				if (dosend(fd, sendstr) < 0)
					return(-1);
			} else {
				return(-1);
			}
		} else {
			return(-1);
		}
	}
	return(0);
}

static int
parse(char **chstrpp, char **expp, char **sndpp,
	char **subexpp, char **subsndpp)
{
	char *s, *p;
	int first;

	*subexpp = NULL;
	*subsndpp = NULL;
	*expp = NULL;
	*sndpp = NULL;
	s = *chstrpp;
	while (isspace(*s))
		s++;
	if (*s == '\0')
		return(0);	/* done */
	*expp = s;
	while ((*s != '\0') && !isspace(*s))
		s++;
	if (*s == '\0') {
		*chstrpp = s;
	} else {
		*chstrpp = s + 1;
		*s = '\0';
	}
	first = 1;
	for (p = *expp; *p != '\0'; p++) {
		if (*p == '-') {
			*p = '\0';
			if (first) {
				*subsndpp = p + 1;
				first = 0;
			} else {
				*subexpp = p + 1;
			}
		}
	}
	if (*++s == '\0')
		return(1);
	*sndpp = s;
	while ((*s != '\0') && !isspace(*s))
		s++;
	if (*s == '\0') {
		*chstrpp = s;
	} else {
		*chstrpp = s + 1;
		*s = '\0';
	}
	return(1);
}

static int
doexpect(int fd, const char *s)
{
	int i, n, len;
	char ebuf[CHATBUFSZ];

	len = strlen(s);
	if (len == 0)	/* null string */
		return(1);
	if ((len == 2) && (s[0] == '"') &&
	  (s[1] == '"')) /* null string */
		return(1);
	errno = 0;
	alarm(60);
	n = 0;
	for (i = 0; i < CHATBUFSZ; i++) {
		if (read(fd, &ebuf[i], 1) != 1) {
			alarm(0);
			return(0);
		}
		if (i < (len-1))
			continue;
		if (strncmp(s, &ebuf[n], len) == 0) {
			alarm(0);
			return(1);
		}
		n++;
		if (errno == EINTR)
			return(0);
	}
	alarm(0);
	return(0);
}

static int
dosend(int fd, const char *s)
{
	int docr;
	uchar_t n;

	docr = 1;
	while (*s != '\0') {
		switch (*s) {
		case '\\':
			s++;
			switch (*s) {
			case 'b':	/* send a backspace */
				if (write(fd, "\b", 1) != 1)
					return(-1);
				break;

			case 'c':	/* don't send a carriage return */
				if (*(s + 1) == '\0')
					docr = 0;
				break;

			case 'd':	/* delay 1 second */
				sleep(1);
				break;

			case 'K':	/* send a break */
				tcsendbreak(fd, 0);
				break;

			case 'n':	/* send a newline */
				if (write(fd, "\n", 1) != 1)
					return(-1);
				break;

			case 'N':	/* send a null */
				if (write(fd, "\0", 1) != 1)
					return(-1);
				break;

			case 'p':	/* pause for part of a second */
				poll(NULL, 0, 300);	/* 300 ms */
				break;

			case 'r':	/* send a carriage return */
				if (write(fd, "\r", 1) != 1)
					return(-1);
				break;

			case 's':	/* send a space */
				if (write(fd, " ", 1) != 1)
					return(-1);
				break;

			case 't':	/* send a tab */
				if (write(fd, "\t", 1) != 1)
					return(-1);
				break;

			case '0':	/* send a byte */
			case '1':	/* with the given */
			case '2':	/* octal value */
			case '3':
				if (isoctal(*(s+1)) && isoctal(*(s+2))) {
					n = tooctal(0, *s++);
					n = tooctal(n, *s++);
					n = tooctal(n, *s);
					if (write(fd, &n, 1) != 1)
						return(-1);
					break;
				}
				/* else fall through */

			default:	/* send the escaped character */
				if (write(fd, s, 1) != 1)
					return(-1);
			}
			s++;
			break;

		default:	/* send the character */
			if (write(fd, s, 1) != 1)
				return(-1);
			s++;
		}
	}
	if (docr) {
		if (write(fd, "\r", 1) != 1)
			return(-1);
	}
	return(0);
}

static void
nop(int signo)
{
	errno = EINTR;
}
